﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Reflection;
using System.Threading;
using System.Threading.Tasks;

namespace cloud.insecurity.http
{
    public class HttpServer
    {
        private static log4net.ILog Log
        {
            get
            {
                return log4net.LogManager.GetLogger(typeof(HttpServer));
            }
        }

        private readonly HttpListener Listener = new HttpListener();
        private readonly static HttpServer Instance = new HttpServer();
        private readonly Tuple<MethodInfo, HttpRequestHandler>[] Handlers;

        public bool Running
        {
            get
            {
                return Listener.IsListening;
            }
            
        }
        
        public static HttpServer Server
        {
            get { return Instance; }
        }

        private HttpServer()
        {
            Handlers = AppDomain.CurrentDomain.GetAssemblies()
                .Distinct()
                .SelectMany(a =>
                {
                    Type[] types;
                    try
                    {
                        types = a.GetTypes();
                    }
                    catch (ReflectionTypeLoadException ex)
                    {
                        HttpServer.Log.Info("lazy load ", ex);
                        types = ex.Types;
                    }
                    
                    return types;
                })
                .Distinct()
                .SelectMany(m => m.GetMethods())
                .Distinct()
                .Where(m => m.GetCustomAttribute(typeof(HttpRequestHandler)) != null)
                .Select(m => new Tuple<MethodInfo, HttpRequestHandler>(m,
                    m.GetCustomAttribute<HttpRequestHandler>())).ToArray();
        }

        public void Stop()
        {
            Listener.Close();
        }

        private IEnumerable<Task<HttpListenerContext>> HandleRequests()
        {
            while (Listener.IsListening)
            {
                yield return Listener.GetContextAsync();
            }
        }
        public async Task<Task> Start(string[] listeners)
        {
            listeners.ToList().ForEach(l => Listener.Prefixes.Add(l));
            
            Listener.Start();
            
            HandleRequests()
                .AsParallel()
                .WithDegreeOfParallelism(8)
                .Select(t =>
                {
                    return t.ContinueWith(task =>
                    {
                        var ctx = task.Result as HttpListenerContext;

                        try
                        {
                            HandleContext(ctx).SendResponse();
                        }
                        catch (NotImplementedException ex)
                        {
                            DefaultHttpResponse.New(ctx)
                                .NotImplemented()
                                .SendResponse();
                        }
                        catch (MissingMethodException ex)
                        {
                            Log.Warn("not found", ex);
                            DefaultHttpResponse.New(ctx)
                                .NotFound()
                                .SendResponse();
                        }
                        catch (Exception ex)
                        {
                            DefaultHttpResponse.New(ctx)
                                .InternalServerError()
                                .SendResponse();
                        }
                    });
                }).ForAll(t => t.Wait());
            
            return Task.CompletedTask;
        }

        private IHttpResponse HandleContext(HttpListenerContext ctx)
        {
            Tuple<MethodInfo, HttpRequestHandler> handler = null;
            IHttpResponse result = null;
            try
            {
                handler = Handlers
                    .Single(t =>
                    {
                        IHttpRequest.RequestMethod method;
                        
                        switch (ctx.Request.HttpMethod)
                        {
                            case "GET":
                                method = IHttpRequest.RequestMethod.Get;
                                break;
                            case "POST":
                                method = IHttpRequest.RequestMethod.Post;
                                break;
                            case "PUT":
                                method = IHttpRequest.RequestMethod.Put;
                                break;
                            case "DELETE":
                                method = IHttpRequest.RequestMethod.Delete;
                                break;
                            case "PATCH":
                                method = IHttpRequest.RequestMethod.Patch;
                                break;
                            case "HEAD":
                                method = IHttpRequest.RequestMethod.Head;
                                break;
                            default:
                                method = IHttpRequest.RequestMethod.Unknown;
                                break;
                        }
                        return t.Item2.Method == method && t.Item2.Uri == ctx.Request.Url?.AbsolutePath;
                    });

                result = (IHttpResponse) handler.Item1.Invoke(
                    handler.Item1.DeclaringType,
                    new object?[]
                    {
                        Activator.CreateInstance(handler.Item1
                                .GetParameters()
                                .Single(p =>
                                    p.ParameterType == typeof(IHttpRequest)).ParameterType,
                            BindingFlags.Static,
                            new object[]
                            {
                                ctx,
                            }),
                    });
            }
            catch (InvalidOperationException ex)
            {
                throw new MissingMethodException("not found", ex);
            }
            
            return result;
        }
    }
}